// test program for BCondVar, structured similarly to a problematic
// case in the BeOS port of Kaffe

#include "condvar.h"
#include <kernel/OS.h>
#include <stdio.h>

// global condition variable and runmode
be_cond_t gcman = BE_COND_INITIALIZER;
be_mutex_t mutex = BE_MUTEX_INITIALIZER;
int runmode = 0;

// thread 1 -- garbage collection daemon
int32 gcdaemon(void*)
{
	be_mutex_lock(&mutex);

	for (;; runmode = 0, fprintf(stderr, "+ broadcasting gcman\n"), be_cond_broadcast(&gcman), fprintf(stderr, "+ sent broadcast\n")) { 
		while (runmode == 0) {
			fprintf(stderr, "+ waiting on gcman\n");
			be_cond_wait(&gcman, &mutex);
			fprintf(stderr, "+ got signal\n");
		} 
		/* Do garbage collection stuff */ 
		fprintf(stderr, "+ *** doing gc here ***\n");
	} 

	return 0;
}

// thread 2 -- tight "System.gc()" loop

void systemgc()
{
	fprintf(stderr, "- locking mutex\n");
	be_mutex_lock(&mutex);
	fprintf(stderr, "- mutex locked\n");
	if (runmode == 0) { 
		runmode = 1; 
		fprintf(stderr, "- signalling gcman\n");
		be_cond_signal(&gcman);
		fprintf(stderr, "- sent signal\n");
	} 
	while (runmode != 0) { 
		fprintf(stderr, "- waiting on gcman\n");
		be_cond_wait(&gcman, &mutex);
		fprintf(stderr, "- got signal\n");
	} 
	be_mutex_unlock(&mutex);
}

int32 systemgcthread(void*)
{
	while (true)
	{
		fprintf(stderr, "- about to call systemgc()\n");
		systemgc();
		fprintf(stderr, "- done calling systemgc()\n");
	}
	return 0;
}

int main(int, char**)
{
	// spawn the gcdaemon thread
	thread_id t1 = ::spawn_thread(gcdaemon, "gcdaemon", B_NORMAL_PRIORITY, NULL);
	// spawn the systemgc loop thread
	thread_id t2 = ::spawn_thread(systemgcthread, "system.gc", B_NORMAL_PRIORITY, NULL);
	// start them both running
	::resume_thread(t1);
	::resume_thread(t2);

	// infinite loop until we're killed
	while (true) snooze(50000);

	return 0;
}
