/* conditional variables for pthreads.... */
#include <stdlib.h>
#include "pthread.h"
#include "pthread_private.h"
#include <stdio.h>

static inline void cond_queue_remove(pthread_cond_t , pthread_t);
static inline void cond_queue_enq(pthread_cond_t , pthread_t);
static inline pthread_t cond_queue_deq(pthread_cond_t);

/*
	allen reese
	allen@kindacool.com
	07/28/00
	adding debugging.  fixup for R5.
	
	allen reese
	allen@driversoft.com
	08/26/99
	
	added init var to pthread_cond_t.
	added support for init to pthread_cond_init
	added delete_sem to pthread_cond_destroy
*/

#include "pthread_debug.h"

int pthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *cond_attr)
{
    enum pthread_cond_type  type;
    pthread_cond_t          bcond;
    int rv = 0;
    
    if (cond == NULL) {
    	pdbgprint ("%s:cond==0x%x,*cond==0x%x\n",__PRETTY_FUNCTION__,cond,(cond)?*cond:0);
    	return EINVAL;
    }
    
    if ((*cond) && (*cond)->init==9999) { 
		/* oops, this condv has already been inited. */
    		pdbgprint ("%s::cond var already inited\n",__PRETTY_FUNCTION__);
    		return EALREADY;
    	}
    
    /* Now we check what was passed in... */
    if (cond_attr != NULL && *cond_attr != NULL){
        /* we were passed a conditional attribute so we get our type
           from that */
        type = (*cond_attr)->c_type;
    } else {
        type = COND_TYPE_FAST; /* the default */
    }
    
    if (type != COND_TYPE_FAST) {
		pdbgprint ("fast init is the only type of condvar supported\n");
        return EINVAL;
    }
    
    if ((bcond = (pthread_cond_t)malloc(sizeof(struct pthread_cond))) == NULL) {
        rv = ENOMEM;
    } else {
        TAILQ_INIT(&bcond->c_queue);
        bcond->c_flags |= COND_FLAGS_INITED;
        bcond->c_type = type;
        bcond->init = 9999;
        _pthread_benaphore_init(&bcond->lock);
        *cond = bcond;
    }
    return (rv);
}

int pthread_cond_destroy(pthread_cond_t * cond) {
    int rv = 0;
    
    if ((cond == NULL) || ((*cond)==NULL)) {
    	pdbgprint ("%s:cond==0x%x,*cond==0x%x\n",__PRETTY_FUNCTION__,cond,(cond)?*cond:0);
   		return EINVAL;
    }
/*    if ((*cond)->init==9999) { 
		/ * oops, this condv has already been inited. ???????-areese* /
    		pdbgprint ("%s::cond var already inited\n",__PRETTY_FUNCTION__);
    		return EALREADY;
    	}
    */
    /* first lock the structure */
    _pthread_benaphore_lock(&(*cond)->lock);
    /* delete the semaphore */
    delete_sem ((*cond)->lock.sem);
    /* now free it !! yikes is this safe?? */
    free ((*cond));
    *cond = NULL;
    return rv;
}

int pthread_cond_wait(pthread_cond_t * cond, pthread_mutex_t *mutex) {
    int rv = 0;
	pthread_t me;

    if ((cond == NULL) || ((*cond)==NULL)) {
    	pdbgprint ("%s:cond==0x%x,*cond==0x%x\n",__PRETTY_FUNCTION__,cond,(cond)?*cond:0);
		return EINVAL;
    }

	me = _find_self();
	if (me == NULL){
    		pdbgprint ("me==NULL\n");
    		return -1;
    }

    if (*cond == NULL || (*cond)->init != 9999) {
    	pdbgprint ("*cond==0x%x\n",(*cond)?*cond:0);
		return errno;
    }

	if ((*cond)->c_type != COND_TYPE_FAST){
		pdbgprint ("fast init is the only type of condvar supported\n");
        return EINVAL;
    }
    
    _pthread_benaphore_lock(&(*cond)->lock);

	if (mutex == NULL || 
		(((*cond)->c_mutex != NULL) && /* we're already using a mutex */
		((*cond)->c_mutex != *mutex))) {
		    _pthread_benaphore_unlock(&(*cond)->lock);
		    return EINVAL;
	}
	
	cond_queue_enq(*cond, me);
	(*cond)->c_mutex = *mutex;
	
	if ((rv = pthread_mutex_unlock(mutex)) != 0) { /* failed to lock */
		cond_queue_remove(*cond, me);
		if (TAILQ_FIRST(&(*cond)->c_queue) == NULL)
			(*cond)->c_mutex = NULL;
		_pthread_benaphore_unlock(&(*cond)->lock);
	} else {
		rv = pthread_mutex_lock(mutex);
		_pthread_benaphore_unlock(&(*cond)->lock);
		/* now suspend the thread */
		suspend_thread(find_thread(NULL));
	}			
    return rv;
}

int pthread_cond_timedwait(pthread_cond_t * cond, pthread_mutex_t *mutex,
				const struct timespec * abstime) {
    int rv = 0;
	pthread_t me;

    if (cond == NULL) {
    	pdbgprint ("%s::passed NULL cond\n",__PRETTY_FUNCTION__);
    	return EINVAL;
    }

	me = _find_self();
	if (me == NULL) {
    	pdbgprint ("%s::unable to findself()\n",__PRETTY_FUNCTION__);
    	return -1;
    }

    if ((*cond == NULL) || ((*cond)->init!=9999)) {
    	pdbgprint ("%s::*cond: 0x%x *cond->init: 0x%dx\n",__PRETTY_FUNCTION__,(*cond)?*cond:0,(*cond)?(*cond)->init:0);
    	return errno;
	};
	
	if ((*cond)->c_type != COND_TYPE_FAST) {
			pdbgprint ("%s::fast init is the only type of condvar supported\n",__PRETTY_FUNCTION__);
			return EINVAL;
	}
    _pthread_benaphore_lock(&(*cond)->lock);

	if (mutex == NULL || 
		(((*cond)->c_mutex != NULL) && /* we're already using a mutex */
		((*cond)->c_mutex != *mutex))) {
			pdbgprint ("%s:mutex problems\n",__PRETTY_FUNCTION__);
		    _pthread_benaphore_unlock(&(*cond)->lock);
		    return EINVAL;
	}
	
	cond_queue_enq(*cond, me);
	(*cond)->c_mutex = *mutex;
	
	if ((rv = pthread_mutex_unlock(mutex)) != 0) { /* failed to lock */
		cond_queue_remove(*cond, me);
		if (TAILQ_FIRST(&(*cond)->c_queue) == NULL)
			(*cond)->c_mutex = NULL;
		_pthread_benaphore_unlock(&(*cond)->lock);
	} else {
		rv = pthread_mutex_lock(mutex);
		_pthread_benaphore_unlock(&(*cond)->lock);
		/* now we snooze for the given time... */
		rv = snooze(abstime->tv_sec * 1000);
	}			
    return rv;
}

int pthread_cond_signal(pthread_cond_t * cond) {
	pthread_t thrd;

    if ((cond == NULL) || (*cond==NULL)) {
    	pdbgprint ("%s:cond==0x%x,*cond==0x%x\n",__PRETTY_FUNCTION__,cond,(cond)?*cond:0);
    	return EINVAL;
    }
	
	if ((*cond)->c_type != COND_TYPE_FAST) {
		pdbgprint ("%s::fast init is the only type of condvar supported\n",__PRETTY_FUNCTION__);
		return EINVAL;
	}

	_pthread_benaphore_lock(&(*cond)->lock);
	
	/* find the first thread to be notified... */
	while (((thrd = cond_queue_deq(*cond)) != NULL ) &&
			(thrd->state != PS_RUNNING)) {}
	if (thrd != NULL) {
		/* OK it's a valid pthread */
		PTHREAD_NEW_STATE(thrd, PS_RUNNING);
		resume_thread(thrd->tid);
	}
	if (TAILQ_FIRST(&(*cond)->c_queue) == NULL)
		(*cond)->c_mutex = NULL;
	_pthread_benaphore_unlock(&(*cond)->lock);
	return 0;
}

int pthread_cond_broadcast(pthread_cond_t * cond) {
	pthread_t thrd;
	
    if ((cond == NULL) || (*cond==NULL)) {
    	pdbgprint ("%s:cond==0x%x,*cond==0x%x\n",__PRETTY_FUNCTION__,cond,(cond)?*cond:0);
    	return EINVAL;
    }
    	
	if ((*cond)->c_type != COND_TYPE_FAST) {
		pdbgprint ("%s::fast init is the only type of condvar supported\n",__PRETTY_FUNCTION__);
		return EINVAL;
	}

	_pthread_benaphore_lock(&(*cond)->lock);
	while ((thrd = cond_queue_deq(*cond)) != NULL) {
		if (thrd->state != PS_RUNNING) { /* bring it back to life ! */
			PTHREAD_NEW_STATE(thrd, PS_RUNNING);
			resume_thread(thrd->tid);
		}
	}
	(*cond)->c_mutex = NULL;
	_pthread_benaphore_unlock(&(*cond)->lock);
	return 0;
}
	
/*********************************/
/* conditional queue routines... */
/*********************************/
static inline pthread_t cond_queue_deq(pthread_cond_t cond) {
	pthread_t pthread;
	
	if ((pthread = TAILQ_FIRST(&cond->c_queue)) != NULL) {
		TAILQ_REMOVE(&cond->c_queue, pthread, qe);
		pthread->flags &= ~PTHREAD_FLAGS_QUEUED;
	}
	return (pthread);
}

static inline void cond_queue_remove(pthread_cond_t cond, pthread_t pthread) {
	if (pthread->flags & PTHREAD_FLAGS_QUEUED) {
		TAILQ_REMOVE(&cond->c_queue, pthread, qe);
		pthread->flags &= ~ PTHREAD_FLAGS_QUEUED;
	}
}

static inline void cond_queue_enq(pthread_cond_t cond, pthread_t pthread) {
	pthread_t tid = TAILQ_LAST(&cond->c_queue, cond_head);
	if ((tid == NULL) || (pthread->active_priority <= tid->active_priority))
		TAILQ_INSERT_TAIL(&cond->c_queue, pthread, qe);
	else {
		tid = TAILQ_FIRST(&cond->c_queue);
		while (pthread->active_priority <= tid->active_priority)
			tid = TAILQ_NEXT(tid, qe);
		TAILQ_INSERT_BEFORE(tid, pthread, qe);
	}
	pthread->flags |= PTHREAD_FLAGS_QUEUED;
}

