#include "pthread.h"
#include "pthread_private.h"
#include <errno.h>
#include <stdio.h>

int pthread_mutex_init(pthread_mutex_t *mtx, const pthread_mutexattr_t *attr )
{
	pthread_mutex_t	new_mtx;

	if (mtx == NULL)
		return EINVAL;
	
	if ((new_mtx = malloc(sizeof(struct pthread_mutex))) == NULL)
		return ENOMEM;

	new_mtx->ben = 0;	
    if ((new_mtx->sem = create_sem(0,"pthread_mutex")) <0) 
        return -1;
    *mtx = new_mtx;
    return 0;
}

int pthread_mutex_lock(pthread_mutex_t *mtx)
{
    if (atomic_add(&(*mtx)->ben,1)>0) {
    	errno=acquire_sem((*mtx)->sem);
        if ( errno!= B_NO_ERROR){
            atomic_add(&(*mtx)->ben,-1);
            return errno;
        }
    }
    return 0;
}

int pthread_mutex_unlock(pthread_mutex_t *mtx)
{
    if (atomic_add(&(*mtx)->ben,-1)>1) {
    	errno=release_sem((*mtx)->sem);
        if (errno != B_NO_ERROR){
            atomic_add(&(*mtx)->ben,1);
            return errno;
        }
    }
    return 0;
}

int pthread_mutex_destroy(pthread_mutex_t *mtx)
{
    delete_sem((*mtx)->sem);
    free ((*mtx));
    return 0;
}
