/* priority queue functions */
#include "pthread.h"
#include "pthread_private.h"

static void pq_insert_prio_list(pq_queue_t *pq, int prio);

int _pq_init(pq_queue_t *pq, int minprio, int maxprio) {
	int i;
	int prioslots = maxprio - minprio + 1;
	
	if (pq == NULL)
		return -1;
	
	if ((pq->pq_lists=(pq_list_t *)malloc(sizeof(pq_list_t) * prioslots))==NULL)
		return (-1);
	for (i=0;i < prioslots; i++) {
		TAILQ_INIT(&pq->pq_lists[i].pl_head);
		pq->pq_lists[i].pl_prio = i;
		pq->pq_lists[i].pl_queued = 0;
	}
	
	TAILQ_INIT(&pq->pq_queue);
	pq->pq_size = prioslots;
	
	return 0;
}

void _pq_remove(pq_queue_t *pq, pthread_t pthread) {
	int prio = pthread->active_priority;
	TAILQ_REMOVE(&pq->pq_lists[prio].pl_head, pthread, pqe);
}

void _pq_insert_head(pq_queue_t *pq, pthread_t pthread) {
	int prio = pthread->active_priority;
	
	TAILQ_INSERT_HEAD(&pq->pq_lists[prio].pl_head, pthread, pqe);
	if (pq->pq_lists[prio].pl_queued == 0)
		pq_insert_prio_list(pq, prio);
}

void _pq_insert_tail(pq_queue_t *pq, pthread_t pthread) {
	int prio = pthread->active_priority;
	
	TAILQ_INSERT_TAIL(&pq->pq_lists[prio].pl_head, pthread, pqe);
	if (pq->pq_lists[prio].pl_queued == 0)
		pq_insert_prio_list(pq, prio);
}

pthread_t _pq_first(pq_queue_t *pq) {
	pq_list_t *pql;
	pthread_t pthread = NULL;
	
	while (((pql = TAILQ_FIRST(&pq->pq_queue)) != NULL) && (pthread == NULL)) {
		if ((pthread = TAILQ_FIRST(&pql->pl_head)) == NULL) {
			/* the priority list is empty..remove it ! */
			TAILQ_REMOVE(&pq->pq_queue, pql, pl_link);
			pql->pl_queued = 0;
		}
	}
	return (pthread);
}

static void pq_insert_prio_list(pq_queue_t *pq, int prio) {
	pq_list_t *pql;
	/* find place in queue for list in descending order of priority */
	pql = TAILQ_FIRST(&pq->pq_queue);
	while ((pql != NULL) && (pql->pl_prio > prio))
		pql = TAILQ_NEXT(pql, pl_link);
	
	if (pql == NULL) 
		TAILQ_INSERT_TAIL(&pq->pq_queue, &pq->pq_lists[prio], pl_link);
	else 
		TAILQ_INSERT_BEFORE(pql, &pq->pq_lists[prio], pl_link);
		
	pq->pq_lists[prio].pl_queued = 1; /* hey we're in the queue! */
}

