/* This file is part of the beginnings of a pthread library for BeOS */

/* TODO : once it's in it's own namespace rename pthread to pthread */

#ifndef PTHREAD_PRIVATE_H
#define PTHREAD_PRIVATE_H

#include <kernel/OS.h>
#include <signal.h>
#include <malloc.h>
#include <errno.h>
#include "pthread_q.h"

/* beos has spinlock but it's nt meant ot be used for this sort of thing,
   so instead we'll use a benaphore :-( */

typedef struct {
    sem_id    sem;
    int32     ben;
}_pthread_benaphore;



/* Constants */
#define PTHREAD_DEFAULT_PRIORITY	10
#define PTHREAD_CREATE_RUNNING		0
#define PTHREAD_CREATE_SUSPENDED	1
#define PTHREAD_FLAGS_PRIVATE		0x0001
#define PTHREAD_EXITING				0x0002
#define PTHREAD_FLAGS_QUEUED		0x0004
#define PTHREAD_FLAGS_TRACE			0x0008
#define COND_FLAGS_PRIVATE  0x01
#define COND_FLAGS_INITED   0x02
#define COND_FLAGS_BUSY     0x04

#define PTHREAD_MIN_PRIORITY	0
#define PTHREAD_MAX_PRIORITY	20 /* arbitrary number for BeOS */

enum pthread_state {
	PS_RUNNING,
	PS_SIGTHREAD,
	PS_MUTEX_WAIT,
	PS_COND_WAIT,
	PS_FDLR_WAIT,
	PS_FDLW_WAIT,
	PS_FDR_WAIT,
	PS_FDW_WAIT,
	PS_FILE_WAIT,
	PS_SELECT_WAIT,
	PS_SLEEP_WAIT,
	PS_WAIT_WAIT,
	PS_SIGSUSPEND,
	PS_SIGWAIT,
	PS_SPINBLOCK,
	PS_JOIN,
	PS_SUSPENDED,
	PS_DEAD,
	PS_DEADLOCK,
	PS_STATE_MAX
};

enum pthread_cond_type {
    COND_TYPE_FAST,
    COND_TYPE_MAX
};

/* structures... */

struct pthread_attr {
	int			priority;
	int 		suspend;
	int 		flags;
	void 		*arg_attr;
};


/* a thread queue structure */
struct pthread_queue {
    struct pthread		*q_next;
    struct pthread		*q_last;
    void                *q_data;
};

struct pthread_cond {
	int init;
    enum pthread_cond_type  c_type;
    TAILQ_HEAD(cond_head, pthread)    c_queue;
    pthread_mutex_t			c_mutex; /* pointer to mutex we're using... */
    void                    *c_data;
    long                    c_flags;
    _pthread_benaphore      lock;
};

struct pthread_cond_attr {
    enum pthread_cond_type  c_type;
    long                    c_flags;
};

struct pthread_mutex {
    enum pthread_mutextype  m_type;
    struct pthread_queue    m_queue;
    struct pthread          *owner;
    long                    m_flags;
    sem_id                  sem;
    int32                   ben;
    /* thanx to chris tate <ctate@be.com> for the init taste test.  lets us return EALREADY */
    int32					init;
};

struct pthread_mutex_attr {
    enum pthread_mutextype  m_type;
    long                    m_flags;
};

/* finally the pthread structure!! */
struct pthread {
    thread_id tid;
    struct pthread      	*nxt;
    struct pthread      	*qnxt;
    struct pthread_queue 	*queue;

	int timeout;    
	
	sigset_t sigmask;
	enum pthread_state		state;
	
	/* queue of threads waiting to join this one... */
	struct pthread_queue	join_queue;
	
	void *args; /* args passed to thread */
	struct pthread_attr attr; /* pthread_attr */	    
    int32 (*start_routine) (void *); /* the start_routine pointer */
	int32 base_priority;
	int32 active_priority;
	
	TAILQ_ENTRY(pthread)	pqe;
	TAILQ_ENTRY(pthread)	qe;
	
	int flags;
#define PTHREAD_FLAGS_PRIVATE		0x0001
#define PTHREAD_EXITING				0x0002
#define PTHREAD_FLAGS_QUEUED		0x0004
#define PTHREAD_FLAGS_TRACE			0x0008

};

/* macros */

#define PTHREAD_MUTEX_STATIC_INITIALIZER \
	{ MUTEX_TYPE_FAST, PTHREAD_QUEUE_INITIALIZER, \
	NULL, MUTEX_FLAGS_INITED ,0 ,0}
	


/* Global variables */
int readyq_initialized;
pq_queue_t	_readyq;
TAILQ_HEAD(, pthread) _waitingq;
struct pthread * _thread_link_list;

/* private defines */
#define MUTEX_FLAGS_PRIVATE  0x01
#define MUTEX_FLAGS_INITED   0x02
#define MUTEX_FLAGS_BUSY     0x04

/* functions... */
/* private */
const inline void _pthread_queue_init(struct pthread_queue *q);
const inline void _pthread_benaphore_init(_pthread_benaphore *);
const inline void _pthread_benaphore_lock(_pthread_benaphore *);
const inline void _pthread_benaphore_unlock(_pthread_benaphore *);

int _pq_init(struct pq_queue *pq, int, int);
void _pq_remove(struct pq_queue *pq, struct pthread *);
void _pq_insert_head(struct pq_queue *pq, struct pthread *);
void _pq_insert_tail(struct pq_queue *pq, struct pthread *);
void _thread_queue_init(struct pthread_queue *);
void _thread_queue_enq(struct pthread_queue *, struct pthread *);
struct pthread* _thread_queue_deq(struct pthread_queue *);
int _thread_queue_remove(struct pthread_queue *, struct pthread *);
void _init_globals();

void _lock_thread_list();
void _unlock_thread_list();
struct pthread * _find_self();
int _find_thread(struct pthread *);


#endif
