/* TAILQ definitions... */

/* these definitions are used in a lot of places! */


#define TAILQ_HEAD(name, type) \
struct name {\
	struct type *tqh_first; /* the first element */ \
	struct type **tqh_last; /* addr of last next element */ \
}
#define TAILQ_HEAD_INITIALIZER(head) \
{ NULL, &(head).tqh_first }

#define TAILQ_ENTRY(type) \
struct { \
	struct type *tqe_next; \
	struct type **tqe_prev; \
}

#define TAILQ_EMPTY (head) ((head)->tqh_first == NULL)

#define TAILQ_FOREACH(var, head, field) \
	for (var = TAILQ_FIRST(head);var; var = TAILQ_NEXT(var, field))

#define TAILQ_FIRST(head) ((head)->tqh_first)

#define TAILQ_LAST(head, headname) \
	(*(((struct headname*)((head)->tqh_last))->tqh_last))

#define TAILQ_NEXT(elm, field) ((elm)->field.tqe_next)

#define TAILQ_PREV(elm, headname, field) \
	(*(((struct headname *)((elm)->field.tqe_prev))->tqh_last))

#define TAILQ_INIT(head) do { \
	(head)->tqh_first = NULL; \
	(head)->tqh_last = &(head)->tqh_first; \
} while (0)

#define TAILQ_INSERT_HEAD(head, elm, field) do { \
	if (((elm)->field.tqe_next = (head)->tqh_first) != NULL)\
		(head)->tqh_first->field.tqe_prev = \
			&(elm)->field.tqe_next;\
	else  \
		(head)->tqh_last = &(elm)->field.tqe_next; \
	(head)->tqh_first = (elm); \
	(elm)->field.tqe_prev = &(head)->tqh_first; \
} while (0)

#define TAILQ_INSERT_TAIL(head, elm, field) do { \
	(elm)->field.tqe_next = NULL; \
	(elm)->field.tqe_prev = (head)->tqh_last; \
	*(head)->tqh_last = (elm); \
	(head)->tqh_last = &(elm)->field.tqe_next; \
} while (0)

#define TAILQ_INSERT_AFTER(head, listelm, elm, field) do { \
	if (((elm)->field.tqe_next = (listelm)->field.tqe_next) != NULL )\\
		(elm)->field.tqe_next->field.tqe_prev = &(elm)->field.tqe_next; \
	else \
		(head)->tqh_last = &(elm)->field.tqe_next; \
	(listelm)->field.tqe_next = (elm); \
	(elm)->field.tqe_prev = &(listelm)->field.tqe_next; \
} while (0)

#define TAILQ_INSERT_BEFORE(listelm, elm, field) do { \
	(elm)->field.tqe_prev = (listelm)->field.tqe_prev; \
	(elm)->field.tqe_next = (listelm); \
	*(listelm)->field.tqe_prev = (elm); \
	(listelm)->field.tqe_prev = &(elm)->field.tqe_next; \
} while (0)

#define TAILQ_REMOVE(head, elm, field) do { \
	if (((elm)->field.tqe_next) != NULL) \
		(elm)->field.tqe_next->field.tqe_prev = (elm)->field.tqe_prev; \
	else \
		(head)->tqh_last = (elm)->field.tqe_prev; \
	*(elm)->field.tqe_prev = (elm)->field.tqe_next; \
} while (0)

/* priority queue macros */
#define PTHREAD_PRIOQ_INSERT_HEAD(thrd)	_pq_insert_head(&_readyq, thrd);
#define PTHREAD_PRIOQ_INSERT_TAIL(thrd) _pq_insert_tail(&_readyq, thrd);
#define PTHREAD_PRIOQ_REMOVE(thrd)		_pq_remove(&_readyq, thrd);
#define PTHREAD_PRIOQ_FIRST				_pq_first(&_readyq);

/* waiting queue macros */
#define PTHREAD_WAITQ_INSERT(thrd)	TAILQ_INSERT_TAIL(&_waitingq,thrd,pqe);
#define PTHREAD_WAITQ_REMOVE(thrd)	TAILQ_REMOVE(&_waitingq,thrd,pqe);

#define PTHREAD_SET_STATE(thrd, newstate) { (thrd)->state = newstate; }

#define PTHREAD_NEW_STATE(thrd, newstate) { \
	if ((thrd)->state != newstate) { \
		if ((thrd)->state == PS_RUNNING) { \
			PTHREAD_PRIOQ_REMOVE(thrd); \
			PTHREAD_WAITQ_INSERT(thrd); \
		} else if (newstate == PS_RUNNING) { \
			PTHREAD_WAITQ_REMOVE(thrd); \
			PTHREAD_PRIOQ_INSERT_TAIL(thrd); \
		} \
	} \
	PTHREAD_SET_STATE(thrd, newstate);\
}

typedef struct pq_list {
	TAILQ_HEAD(, pthread)	pl_head; /* for this priority a list of threads */
	TAILQ_ENTRY(pq_list)	pl_link; /* link to queue of prioity lists */
	int pl_prio; /* priority of this list */
	int pl_queued; /* have I been queued?? */
} pq_list_t;

typedef struct pq_queue {
	TAILQ_HEAD(, pq_list)	pq_queue; /* queue of priority lists */
	pq_list_t				*pq_lists; /* array of all priority lists */
	int 					pq_size; /* number of priority lists */
} pq_queue_t;

#define PTHREAD_QUEUE_INITIALIZER { NULL, NULL, NULL }
#define TAILQ_INITIALIZER { NULL, NULL }

