/* -*- C++ -*- */
/*
	rwlocker.c
	
	Author: Allen Reese	
	allen@driversoft.com

	Created: 02/08/00 21:39:22
*/

/*
	NOTES:
		Queues aren't here yet.  this is still half unwritten.
		
		A API that will be calling these function shas one minor problem.
		It is based upon the Win32 API for read/write locks, and wants to 
		call unlock on a reader or writer lock, and have the library
		figure out whether it is a reader or a writer unlocking it.
*/

#ifndef _RWLOCKER_H_
#include <rwlocker.h>
#endif /*_RWLOCKER_H_*/

int		pthread_rwlock_init_np (pthread_rwlock_t *rwlock,
			const pthread_rwlockattr_t *rwlock_attr)
{
	int ret=0;
	
	/* caller alloc's the struct */
	if (!rwlock) {
		return EINVAL;
	};
	
	if (rwlock==RWLOCK_INIT_FINISH) {
		/* hey, someone passed us an already inited rwlock, puke. */
		return EALREADY;
	}
	
	/* for later debugging.  This will help if something wierd happens. */
	rwlock->init=RWLOCK_INIT_BEGIN;
	
	/* init the rwlock */
	
	readers=0;
	writers=0;
	
	/* init the reader mutex */
	ret=pthread_mutex_init (&reader_mutex,NULL);
	if (ret!=0) {
		return ret;
	};
	
	/* init the writer mutex */
	ret=pthread_mutex_init (&writer_mutex,NULL);
	if (ret!=0) {
		return ret;
	};
		
	/* for later debugging.  This will help if something wierd happens. */
	rwlock->init=RWLOCK_MUTEXES_INITED;
	
	if (rwlock_attr) {
		/* do something with the attributes */
		ret=pthread_rwlockattr_init (rwlock_attr);
		if (ret!=0) {
			return ret;
		};
		
		attr_struct=malloc (sizeof (pthread_rwlockattr_t));
		if (attr_struct=NULL) {
			return ENOMEM;
		};
		
		memcpy (attr_struct,rwlock_attr,sizeof (pthread_rwlockattr_t));
	};

	/* we're done, set the init flag. */
	rwlock->init=RWLOCK_INIT_FINISH;

	return 0;
}

/* the block parametere is not yet implemented */
int		pthread_rwlock_destroy_np (pthread_rwlock_t *rwlock,bool block)
{
	if ((!rwlock) || (rwlock==RWLOCK_INIT_FINISH) ){
		return EINVAL;
	};

	if ((rwlock->readers>0) || (rwlock->writers>0)) {
		/* oops, there is still a reader reference, return an error */
		return EWOULDBLOCK;
	};
	
	/* free the semaphores and such */
	pthread_mutex_destroy (&reader_mutex);
	pthread_mutex_destroy (&writer_mutex);
	
	/* free the attributes struct if it exists */
	pthread_rwlockattr_destroy_np (attr_struct);

	/* caller frees the struct */
	return 0;
}

int		pthread_rwlock_rdlock_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	if (rwlock->writers>0) {
		/* oops, there are writers waiting, or writing. */
		return EWOULDBLOCK;
	} else {
		if (readers<=0) {
			int ret=0;
			/* aquire the reading semaphore. */
			ret=pthread_mutex_trylock (rwlock->reader_mutex);
			
			if (ret!=EOK) {
				return ret;
			};
		};
		atomic_add (rwlock->readers);
	};
	
	return EOK;
}

int		pthread_rwlock_wrlock_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	return EWOULDBLOCK;
}

int		pthread_rwlock_tryrdlock_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	if (rwlock->writers>0) {
		/* oops, there are writers waiting, or writing. */
		return EWOULDBLOCK;
	} else {
		if (readers<=0) {
			int ret=0;
			/* aquire the reading semaphore. */
			ret=pthread_mutex_trylock (rwlock->reader_mutex);
			
			if (ret!=EOK) {
				return ret;
			};
		};
		atomic_add (&(rwlock->readers),1);
	};
	
	return EOK;
}

int		pthread_rwlock_trywrlock_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	return EWOULDBLOCK;
}

/* you must call the approriate unlock accord to the lock you aquired */
int		pthread_rwlock_unlock_reader_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	if (readers>1) {
		atomic_add (&(rwlock->readers),-1)
	} else {
		atomic_add (&(rwlock->readers),-1)
		ret=pthread_mutex_unlock (rwlock->reader_mutex);
		if (ret!=EOK) {
			return ret;
		};
	}
	
	return EOK;
}

int		pthread_rwlock_unlock_writer_np (pthread_rwlock_t *rwlock)
{
	if (rwlock==NULL) {
		return EINVAL;
	};

	if (writers>0) {
		atomic_add (&(rwlock->writers),-1)
		/* we also need to take care of letting the next writer
			unblock and start writing. */
	} else {
		/* if there are no writers, you can't release one. */
		return EINVAL;
	}
	
	return EOK;
}

int		pthread_rwlockattr_init_np (pthread_rwlockattr_t *lockattr)
{
	if (lockattr==NULL) {
		return EINVAL;
	};
	
	lock_attr->pshared=NULL;
	lock_attr->init=RWLOCK_INIT_FINISH;

	return EOK;
}

int		pthread_rwlockattr_getpshared (const pthread_rwlockattr_t *lockattr,
			rwlock_pshared_struct *pshared)
{
	if (lockattr==NULL) {
		return EINVAL;
	};
	
	pshared=lock_attr->pshared;
	
	return EOK;
}
			
int		pthread_rwlockattr_setpshared_np (pthread_rwlockattr_t *lockattr,
			rwlock_pshared_struct *pshared)
{
	if (lockattr==NULL) {
		return EINVAL;
	};
	
	lockattr->pshared=pshared;
	
	return EOK;
}

			
int		pthread_rwlockattr_destroy_np (pthread_rwlockattr_t *lockattr)
{
	if (lockattr==NULL) {
		return EINVAL;
	};
	
	if ((lockattr->pshared) && (lockattr->pshared->refcount<=0)) {
		free (lockattr->pshared);
		free (lockattr);
	} else {
		return EWOULDBLOCK;
	};
	
	return EOK;
}
