/* -*- C++ -*- */
/*
	rwlocker.h
	
	Author: Allen Reese
	allen@driversoft.com
	
	Created: 02/08/00 21:39:22
*/

#ifndef _RWLOCKER_H_
#define _RWLOCKER_H_

/* pthreads headers we need */
#include "pthread.h"
#include "pthread_private.h"

/* system headers we need */
#include <malloc.h>
#include <string.h>

/* rwlock->init values, for taste testing */
#define RWLOCK_INIT_UNINT		0xDEAD
#define RWLOCK_INIT_BEGIN		0xBEEF
#define RWLOCK_INIT_FINISH		0xC001
#define RWLOCK_MUTEXES_INITED	0xFADE

/*
	you must call the approriate unlock accord to the lock you aquired.
	
	Rather than expensive _find_thread() calls to keep track of all of the 
	readers and writers tid's, I decided to force the caller to unlock the 
	same way the locked.  by unlocking a reader or unlocking a writer.
	
	When pthreads uses Thread Local Storage then maybe this will change.
	
	Allen Reese
	allen@driversoft.com
*/

/* the block parametere is not yet implemented */

struct rwlock_pshared_struct
{
	int init=RWLOCK_INIT_UNINT;

	/* this will be incremented by setpshared, and decremented by _destroy */
	int refcount=0;
	void *pshared=NULL;
};

struct pthread_rwlockattr
{
	int init=RWLOCK_INIT_UNINT;
	
	struct rwlock_pshared_struct *pshared;
};

struct pthread_rwlock {
	/* taste test cookie */
	int init=RWLOCK_INIT_UNINT;

	/* number of read locks aquired */
	int readers=0;
	/* number of write locks aquired, should never be >1 or <0 */
	int writers=0;

	/* the first reader aquires this, all subsequent will just inc readers */
	pthread_mutex_t reader_mutex;
	/* each writer will block on this */
	pthread_mutex_t writer_mutex;
	
	
	/* need 2 priority queues, sleeping readers and sleeping writers. */

	/* the attributes struct */
	struct pthread_rwlockattr *attr_struct; 	
	
};


int		pthread_rwlock_init_np (pthread_rwlock_t *rwlock,
			const pthread_rwlockattr_t *lockattr);
int		pthread_rwlock_destroy_np (pthread_rwlock_t *rwlock,bool block=false)

/* try for a blocking read lock, won't return EWOULDBLOCK */
int		pthread_rwlock_rdlock_np (pthread_rwlock_t *rwlock);
/* try for a blocking write lock, won't return EWOULDBLOCK */
int		pthread_rwlock_wrlock_np (pthread_rwlock_t *rwlock);
/* try for a read lock, fail if EWOULDBLOCK */
int		pthread_rwlock_tryrdlock_np (pthread_rwlock_t *rwlock);
/* try for a write lock, fail if EWOULDBLOCK */
int		pthread_rwlock_trywrlock_np (pthread_rwlock_t *rwlock);

/* you must call the approriate unlock accord to the lock you aquired */
int		pthread_rwlock_unlock_reader_np (pthread_rwlock_t *rwlock);
int		pthread_rwlock_unlock_writer_np (pthread_rwlock_t *rwlock);

int		pthread_rwlockattr_init_np (pthread_rwlockattr_t *lockattr);
int		pthread_rwlockattr_getpshared_np (const pthread_rwlockattr_t *lockattr,
			rwlock_pshared_struct *pshared);
int		pthread_rwlockattr_setpshared_np (pthread_rwlockattr_t *lockattr,
			rwlock_pshared_struct *pshared);
int		pthread_rwlockattr_destroy_np (pthread_rwlockattr_t *lockattr);

#endif /*_RWLOCKER_H_*/
